/*
    Cyclops - Windows 95 demo program
    Demonstrates taskbar icon notification
    Copyright (c) 1995-2003 Alessandro Scotti
    http://www.walkofmind.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#define STRICT
#include <windows.h>
#pragma hdrstop
#include <windowsx.h>

#include "tna.h"
#include "cyclops.rh"

/* Global variables */
static CHAR         szAppName[] = "Cyclops95";  // Application name
static HICON        hAppIcon;                   // Application icon
static HMENU        hPopupMenu = NULL;          // Popup menu handle
static HICON        hEyeIcon[10];               // Eye icons
static BOOL         boEnabled = TRUE;           // State
// Tray tooltips
static CHAR         szTipAwake[] = "Cyclops is watching...";
static CHAR         szTipSleep[] = "Zzzz";

/* Constants */
#define TIMERPOLLRATE   100     // Timer poll rate (ms)
#define TIMERIDLETIME   30      // Enter idle after xx WM_TIMER messages
#define MAX_INSTANCES   2       // Max number of instances

/* Tray notification constants */
#define TNID_EYEICON        1                   // Icon identifier
#define TNWM_NOTIFYICON     (WM_USER+0x0100)    // Notification message

/* Popup menu identifiers */
#define IDM_CLOSE           1
#define IDM_STATE           2
#define IDM_ABOUT           3

/*----------------------------------------------------------

    Adds an icon to the taskbar.

    Input:
        hWnd    = handle of window to receive notification messages
        hIcon   = handle of icon to show
        uID     = icon identifier
        lpszTip = tooltip message
    Output:
        result

----------------------------------------------------------*/
BOOL tbarAddIcon( HWND hWnd, UINT uID, HICON hIcon, LPSTR lpszTip )
{
    NOTIFYICONDATA  nid;

    nid.cbSize = sizeof(NOTIFYICONDATA);
    nid.hWnd = hWnd;
    nid.hIcon = hIcon;
    nid.uID = uID;
    nid.uCallbackMessage = TNWM_NOTIFYICON;
    nid.uFlags = NIF_ICON | NIF_MESSAGE;
    if( lpszTip ) {
        strcpy( nid.szTip, lpszTip );
        nid.uFlags |= NIF_TIP;
    }
    else {
        strcpy( nid.szTip, "" );
    }

    return( (BOOL)Shell_NotifyIcon( NIM_ADD, &nid ) );
}

/*----------------------------------------------------------

    Removes an icon from the taskbar.

    Input:
        hWnd    = handle of window to receive notification messages
        uID     = icon identifier
    Output:
        result

----------------------------------------------------------*/
BOOL tbarDelIcon( HWND hWnd, UINT uID )
{
    NOTIFYICONDATA  nid;

    nid.cbSize = sizeof(NOTIFYICONDATA);
    nid.hWnd = hWnd;
    nid.uID = uID;

    return( (BOOL)Shell_NotifyIcon( NIM_DELETE, &nid ) );
}

/*----------------------------------------------------------

    Modifies an icon in the taskbar.

    Input:
        hWnd    = handle of window to receive notification messages
        uID     = icon identifier
        hIcon   = new handle of icon (NULL to keep current)
        lpszTip = new tooltip message (NULL to keep current)
    Output:
        result

----------------------------------------------------------*/
BOOL tbarModifyIcon( HWND hWnd, UINT uID, HICON hIcon, LPSTR lpszTip )
{
    BOOL            boRes;
    NOTIFYICONDATA  nid;

    nid.cbSize = sizeof(NOTIFYICONDATA);
    nid.hWnd = hWnd;
    nid.uID = uID;
    nid.uFlags = 0;
    if( hIcon ) {
        nid.hIcon = hIcon;
        nid.uFlags |= NIF_ICON;
    }
    if( lpszTip ) {
        strcpy( nid.szTip, lpszTip );
        nid.uFlags |= NIF_TIP;
    }

    return( (BOOL)Shell_NotifyIcon( NIM_MODIFY, &nid ) );
}

/*----------------------------------------------------------

    Handle taskbar notification messages.

    Input:
        hWnd    = handle of notification window
        uID     = icon identifier
        wMsg    = mouse message
    Output:
        result (0 if message was handled)

----------------------------------------------------------*/
LRESULT OnNotifyIcon( HWND hWnd, UINT uID, UINT wMsg )
{
    DWORD   dwCurPos;

    switch( uID ) {
        case TNID_EYEICON:
            switch( wMsg ) {
                case WM_LBUTTONDBLCLK:
                    /* Change state (default menu command) */
                    SendMessage( hWnd, WM_COMMAND, IDM_STATE, 0 );
                    break;
                case WM_RBUTTONUP:
                    /* Display popup menu at cursor position, */
                    /* make thread active to track the cursor */
                    SetForegroundWindow( hWnd );
                    dwCurPos = GetMessagePos();
                    TrackPopupMenu( hPopupMenu,
                        TPM_LEFTALIGN | TPM_RIGHTBUTTON,
                        LOWORD(dwCurPos), HIWORD(dwCurPos),
                        0,
                        hWnd,
                        NULL );
                    break;
            }
            break;
    }
    return( 0 );
}

/*----------------------------------------------------------

    Modifies the popup menu according to the value of
    the global boState variable.

----------------------------------------------------------*/
void subSetPopupMenuState( void )
{
    ModifyMenu( hPopupMenu,
        IDM_STATE,
        MF_STRING | MF_BYCOMMAND,
        IDM_STATE,
        (boEnabled ? "Sleep" : "Watch") );
}

/*----------------------------------------------------------

    Handle window messages.

    Input:
        hWnd            = handle of window
        wMsg            = message
        wParam          = message word parameter
        lParam          = message long parameter
    Output:
        result (0 if message was handled)

----------------------------------------------------------*/
LRESULT CALLBACK _export WndProc( HWND hWnd, UINT wMsg,
    WPARAM wParam, LPARAM lParam )
{
    static UINT     uTimer = 0x0001;        // Timer identifier
    static UINT     uLastIdx = 0;
    static UINT     uDelay = 0;
    UINT            uCx, uCy, uIdx;
    POINT           stPt;

    switch( wMsg ) {
        /* Initialize timer on window creation */
        case WM_CREATE:
            SetTimer( hWnd, uTimer, TIMERPOLLRATE, NULL );
            break;
        /* Timer */
        case WM_TIMER:
            if( !boEnabled )
                break;
            GetCursorPos( &stPt );
            uCx = stPt.x / (GetSystemMetrics( SM_CXSCREEN ) / 3);
            if( uCx >= 3 ) uCx = 2;
            uCy = stPt.y / (GetSystemMetrics( SM_CYSCREEN ) / 3);
            if( uCy >= 3 ) uCy = 2;
            uIdx = uCy*3 + uCx + 1;
            if( uIdx != uLastIdx ) {
                /* Modify taskbar icon */
                tbarModifyIcon( hWnd,
                    TNID_EYEICON,
                    hEyeIcon[uIdx],
                    NULL );
                uLastIdx = uIdx;
                uDelay = 0;
            }
            else {
                /* Bump delay count, check for idle time */
                uDelay++;
                if( uDelay == TIMERIDLETIME )
                    tbarModifyIcon( hWnd,
                        TNID_EYEICON,
                        hEyeIcon[0],
                        NULL );
            }
            return( 0 );
        /* Taskbar icon (tray) notification */
        case TNWM_NOTIFYICON:
            return( OnNotifyIcon( hWnd, wParam, (UINT)lParam ) );
        /* Popup menu commands */
        case WM_COMMAND:
            switch( wParam ) {
                /* Display about box */
                case IDM_ABOUT:
                    EnableMenuItem( hPopupMenu,
                        IDM_ABOUT,
                        MF_BYCOMMAND | MF_GRAYED );
                    ShellAbout( hWnd,
                        "",
                        "Cyclops 95, by Alessandro Scotti",
                        hAppIcon );
                    EnableMenuItem( hPopupMenu,
                        IDM_ABOUT,
                        MF_BYCOMMAND | MF_ENABLED );
                    break;
                /* Close application */
                case IDM_CLOSE:
                    DestroyWindow( hWnd );
                    break;
                /* Change eye state */
                case IDM_STATE:
                    if( boEnabled ) {
                        /* Put eye to sleep */
                        boEnabled = FALSE;
                        tbarModifyIcon( hWnd,
                            TNID_EYEICON, hEyeIcon[0], szTipSleep );
                    }
                    else {
                        /* Awake */
                        boEnabled = TRUE;
                        tbarModifyIcon( hWnd,
                            TNID_EYEICON, hEyeIcon[5], szTipAwake );
                        uLastIdx = 0;
                    }
                    subSetPopupMenuState();
                    break;
            }
            return( 0 );
        /* Quit application */
        case WM_DESTROY:
            KillTimer( hWnd, uTimer );
            PostQuitMessage( 0 );
            return( 0 );
    }
    return( DefWindowProc( hWnd, wMsg, wParam, lParam ) );
}

/*----------------------------------------------------------

    Program entry point.

    Input:
        hInstance       = handle of application instance
        hPrevInstance   = handle of previous application instance
        lpszCmd         = command line
        nCmdShow        = show mode
    Output:
        exit code

----------------------------------------------------------*/
#pragma argsused
int PASCAL WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    LPSTR lpszCmd, int nCmdShow )
{
    HWND        hWnd;
    MSG         stMsg;
    WNDCLASS    stWndClass;
    UINT        wTmp;
    HANDLE      hSem;
    DWORD       dwWait;

    /* To avoid too many instances, we must create a synchronization */
    /* object, because for 32-bit applications hPrevInstance is NULL */
    hSem = CreateSemaphore( NULL,   // Create semaphore (will open
        MAX_INSTANCES,              // if a semaphore of the same
        MAX_INSTANCES,              // name already exists)
        "CyclopsSemaphore" );
    if( hSem == NULL )              // Failed to create/open, exit
        return( 0 );
    dwWait = WaitForSingleObject( hSem, 0L );
    if( dwWait != WAIT_OBJECT_0 ) { // Semaphore not signaled or error,
        return( 0 );                // assume too many instances
    }

    /* Set return value to zero */
    stMsg.wParam = 0;

    /* Register window class */
    stWndClass.style            = CS_HREDRAW | CS_VREDRAW;
    stWndClass.lpfnWndProc      = (WNDPROC)WndProc;
    stWndClass.cbClsExtra       = 0;
    stWndClass.cbWndExtra       = 0;
    stWndClass.hInstance        = hInstance;
    stWndClass.hIcon            = (HICON)NULL;
    stWndClass.hCursor          = (HCURSOR)NULL;
    stWndClass.hbrBackground    = (HBRUSH)GetStockObject(NULL_BRUSH);
    stWndClass.lpszMenuName     = NULL;
    stWndClass.lpszClassName    = szAppName;
    if( RegisterClass( &stWndClass ) == 0 )
        goto lblExit;

    /* Create window */
    hWnd = CreateWindow( szAppName,
        "Cyclops",
        WS_POPUP,
        0, 0,
        0, 0,
        NULL,
        NULL,
        hInstance,
        NULL );
    if( hWnd == NULL )
        goto lblExit;

    /* Create popup menu */
    hPopupMenu = CreatePopupMenu();
    if( hPopupMenu == NULL )
        goto lblExit;
    AppendMenu( hPopupMenu, MF_STRING, IDM_STATE, "" );
    AppendMenu( hPopupMenu, MF_STRING | MF_SEPARATOR, 0, NULL );
    AppendMenu( hPopupMenu, MF_STRING, IDM_ABOUT, "About" );
    AppendMenu( hPopupMenu, MF_STRING, IDM_CLOSE, "Exit" );
    subSetPopupMenuState();
    SetMenuDefaultItem( hPopupMenu, IDM_STATE, FALSE );

    /* Load eye icons: icons must be stored in the proper order, */
    /* because we always access them by number */
    hAppIcon = LoadIcon( hInstance, MAKEINTRESOURCE(ICON_MAIN) );
    for( wTmp=0; wTmp<10; wTmp++ )
        /* Use new function to avoid image manipulation */
        hEyeIcon[wTmp] = (HICON)LoadImage( hInstance,
            MAKEINTRESOURCE(ICON_X+wTmp),
            IMAGE_ICON,
            16, 16,
            LR_SHARED );

    /* Add icon to the taskbar */
    tbarAddIcon( hWnd, TNID_EYEICON, hEyeIcon[5], szTipAwake );

    /* Enter main loop */
    while( GetMessage( &stMsg, NULL, 0, 0 ) ) {
        TranslateMessage( &stMsg );
        DispatchMessage( &stMsg );
    }

    /* Free resources */
    tbarDelIcon( hWnd, TNID_EYEICON );
    for( wTmp=0; wTmp<10; wTmp++ )
        DestroyIcon( hEyeIcon[wTmp] );
    DestroyIcon( hAppIcon );
    DestroyMenu( hPopupMenu );

    /* Release semaphore and terminate program */
lblExit:
    ReleaseSemaphore( hSem, 1, NULL );
    return( stMsg.wParam );
}
